/* Copyright 2015 Marvin Beckers <beckersmarvin@gmail.com>
*
* This program is free software: you can redistribute it
* and/or modify it under the terms of the GNU General Public License as
* published by the Free Software Foundation, either version 3 of the
* License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be
* useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
* Public License for more details.
*
* You should have received a copy of the GNU General Public License along
* with this program. If not, see http://www.gnu.org/licenses/.
*/

public class AppCenterCore.Client : Object {
    public signal void updates_available ();
    public signal void tasks_finished ();

    public bool connected { public get; private set; }
    public bool updating_cache { public get; private set; }

    public AppCenterCore.Package os_updates { public get; private set; }

    private Gee.LinkedList<AppCenter.Task> task_list;
    private Gee.LinkedList<AppCenter.Task> task_with_agreement_list;
    private Gee.HashMap<string, AppCenterCore.Package> package_list;
    private AppStream.Pool appstream_pool;
    private GLib.Cancellable interface_cancellable;
    private GLib.DateTime last_cache_update;
    private uint updates_number = 0U;
    private uint update_cache_timeout_id = 0;
    private bool refresh_in_progress = false;

    private const int SECONDS_BETWEEN_REFRESHES = 60*60*24;

    private Client () {
        appstream_pool.get_components ().foreach ((comp) => {
            var package = new AppCenterCore.Package (comp);
            foreach (var pkg_name in comp.get_pkgnames ()) {
                package_list.set (pkg_name, package);
            }
        });
    }

    construct {
        task_list = new Gee.LinkedList<AppCenter.Task> ();
        task_with_agreement_list = new Gee.LinkedList<AppCenter.Task> ();
        package_list = new Gee.HashMap<string, AppCenterCore.Package> (null, null);
        interface_cancellable = new GLib.Cancellable ();
        last_cache_update = null;

        appstream_pool = new AppStream.Pool ();
        try {
            appstream_pool.load ();
        } catch (Error e) {
            error (e.message);
        }

        var os_updates_component = new AppStream.Component ();
        os_updates_component.id = AppCenterCore.Package.OS_UPDATES_ID;
        os_updates_component.name = _("Operating System Updates");
        os_updates_component.summary = _("Updates to system components");
        var icon = new AppStream.Icon ();
        icon.set_name ("distributor-logo");
        icon.set_kind (AppStream.IconKind.STOCK);
        os_updates_component.add_icon (icon);
        os_updates = new AppCenterCore.Package (os_updates_component);
    }

    public bool has_tasks () {
        return !task_list.is_empty;
    }

    private AppCenter.Task request_task (bool requires_user_agreement = true) {
        AppCenter.Task task = new AppCenter.Task ();
        task_list.add (task);
        if (requires_user_agreement) {
            if (task_with_agreement_list.size == 0) {
                Pk.polkit_agent_open ();
            }
            task_with_agreement_list.add (task);
        }
        return task;
    }

    private void release_task (AppCenter.Task task) {
        task_list.remove (task);
        if (task_list.is_empty) {
            tasks_finished ();
            if (task in task_with_agreement_list) {
                task_with_agreement_list.remove (task);
                if (task_with_agreement_list.size == 0) {
                    Pk.polkit_agent_close ();
                }
            }
        }
    }

    public async Pk.Exit install_package (Package package, Pk.ProgressCallback cb, GLib.Cancellable cancellable) throws GLib.Error {
        Pk.Exit exit_status = Pk.Exit.UNKNOWN;
        AppCenter.Task install_task = request_task ();
        AppCenter.Task search_task = request_task ();
        string[] packages_ids = {};
        foreach (var pkg_name in package.component.get_pkgnames ()) {
            packages_ids += pkg_name;
        }

        packages_ids += null;

        try {
            var results = yield search_task.resolve_async (Pk.Bitfield.from_enums (Pk.Filter.NEWEST, Pk.Filter.ARCH), packages_ids, cancellable, () => {});
            packages_ids = {};

            results.get_package_array ().foreach ((package) => {
                packages_ids += package.package_id;
            });

            packages_ids += null;

            results = yield install_task.install_packages_async (packages_ids, cancellable, cb);
            exit_status = results.get_exit_code ();
            if (exit_status != Pk.Exit.SUCCESS) {
                release_task (search_task);
                release_task (install_task);
                throw new GLib.IOError.FAILED (Pk.Exit.enum_to_string (results.get_exit_code ()));
            }
        } catch (Error e) {
            release_task (search_task);
            release_task (install_task);
            throw e;
        }

        release_task (search_task);
        release_task (install_task);
        return exit_status;
    }

    public async Pk.Exit update_package (Package package, Pk.ProgressCallback cb, GLib.Cancellable cancellable) throws GLib.Error {
        Pk.Exit exit_status = Pk.Exit.UNKNOWN;
        SuspendControl sc = new SuspendControl ();
        AppCenter.Task update_task = request_task ();
        string[] packages_ids = {};
        foreach (var pk_package in package.change_information.changes) {
            packages_ids += pk_package.get_id ();
        }
        
        packages_ids += null;

        try {
            sc.inhibit ();
            var results = yield update_task.update_packages_async (packages_ids, cancellable, cb);
            exit_status = results.get_exit_code ();
        } catch (Error e) {
            throw e;
        } finally {
            sc.uninhibit ();
            release_task (update_task);
        }

        if (exit_status != Pk.Exit.SUCCESS) {
            throw new GLib.IOError.FAILED (Pk.Exit.enum_to_string (exit_status));
        } else {
            package.change_information.clear_update_info ();
        }

        yield refresh_updates ();

        return exit_status;
    }

    public async Pk.Exit remove_package (Package package, Pk.ProgressCallback cb, GLib.Cancellable cancellable) throws GLib.Error {
        Pk.Exit exit_status = Pk.Exit.UNKNOWN;
        AppCenter.Task remove_task = request_task ();
        AppCenter.Task search_task = request_task ();
        string[] packages_ids = {};
        foreach (var pkg_name in package.component.get_pkgnames ()) {
            packages_ids += pkg_name;
        }
        packages_ids += null;

        try {
            var results = yield search_task.resolve_async (Pk.Bitfield.from_enums (Pk.Filter.INSTALLED, Pk.Filter.NEWEST), packages_ids, cancellable, () => {});
            packages_ids = {};
            results.get_package_array ().foreach ((package) => {
                packages_ids += package.package_id;
            });

            results = yield remove_task.remove_packages_async (packages_ids, true, true, cancellable, cb);
            exit_status = results.get_exit_code ();
        } catch (Error e) {
            release_task (search_task);
            release_task (remove_task);
            throw e;
        }

        yield refresh_updates ();
        release_task (search_task);
        release_task (remove_task);
        return exit_status;
    }

    public async void get_updates () {
        AppCenter.Task update_task = request_task (false);
        AppCenter.Task details_task = request_task (false);
        try {
            Pk.Results result = yield update_task.get_updates_async (0, interface_cancellable, (t, p) => { });
            string[] packages_array = {};
            result.get_package_array ().foreach ((pk_package) => {
                packages_array += pk_package.get_id ();
            });

            // We need a null to show to PackageKit that it's then end of the array.
            packages_array += null;

            Pk.Results result2 = yield details_task.get_details_async (packages_array , interface_cancellable, (t, p) => { });
            result2.get_details_array ().foreach ((pk_detail) => {
                var pk_package = new Pk.Package ();
                try {
                    pk_package.set_id (pk_detail.get_package_id ());
                    unowned string pkg_name = pk_package.get_name ();
                    var package = package_list.get (pkg_name);
                    if (package == null) {
                        package = os_updates;
                        var pkgnames = os_updates.component.pkgnames;
                        pkgnames += pkg_name;
                        os_updates.component.pkgnames = pkgnames;
                    }

                    package.change_information.changes.add (pk_package);
                    package.change_information.details.add (pk_detail);
                    package.update_state ();
                } catch (Error e) {
                    critical (e.message);
                }
            });
        } catch (Error e) {
            // Error code 19 is for operation canceled.
            if (e.code != 19) {
                critical (e.message);
            }
        }

        release_task (details_task);
        release_task (update_task);
        updates_available ();
    }

    public async Gee.Collection<AppCenterCore.Package> get_installed_applications () {
        var packages = new Gee.TreeSet<AppCenterCore.Package> ();
        var packages_list = yield get_installed_packages ();
        foreach (var pk_package in packages_list) {
            var package = package_list.get (pk_package.get_name ());
            if (package != null) {
                package.installed_packages.add (pk_package);
                package.update_state ();
                packages.add (package);
            }
        }

        return packages;
    }

    public Gee.Collection<AppCenterCore.Package> get_applications_for_category (AppStream.Category category) {
        unowned GLib.GenericArray<AppStream.Component> components = category.get_components ();
        if (components.length == 0) {
            var category_array = new GLib.GenericArray<AppStream.Category> ();
            category_array.add (category);
            AppStream.utils_sort_components_into_categories (appstream_pool.get_components (), category_array, true);
            components = category.get_components ();
        }

        var apps = new Gee.TreeSet<AppCenterCore.Package> ();
        components.foreach ((comp) => {
            apps.add (package_list.get (comp.get_pkgnames ()[0]));
        });

        return apps;
    }

    public Gee.Collection<AppCenterCore.Package> search_applications (string query, AppStream.Category? category) {
        var apps = new Gee.TreeSet<AppCenterCore.Package> ();
        GLib.GenericArray<weak AppStream.Component> comps = appstream_pool.search (query);
        if (category == null) {
            comps.foreach ((comp) => {
                apps.add (package_list.get (comp.get_pkgnames ()[0]));
            });
        } else {
            var cat_packages = get_applications_for_category (category);
            comps.foreach ((comp) => {
                var package = package_list.get (comp.get_pkgnames ()[0]);
                if (package in cat_packages) {
                    apps.add (package);
                }
            });
        }

        return apps;
    }

    public Pk.Package? get_app_package (string application, Pk.Bitfield additional_filters = 0) throws GLib.Error {
        AppCenter.Task packages_task = request_task (false);
        Pk.Package? package = null;
        var filter = Pk.Bitfield.from_enums (Pk.Filter.NEWEST);
        filter |= additional_filters;
        try {
            var results = packages_task.search_names_sync (filter, { application, null }, interface_cancellable, () => {});
            var array = results.get_package_array ();
            if (array.length > 0) {
                package = array.get (0);
            }
        } catch (Error e) {
            release_task (packages_task);
            throw e;
        }

        release_task (packages_task);
        return package;
    }

    private async void refresh_updates () {
        var update_task = new AppCenter.Task ();
        updating_cache = true;

        try {
            Pk.Results result = yield update_task.get_updates_async (0, interface_cancellable, (t, p) => {});
            bool was_empty = updates_number == 0U;
            updates_number = get_package_count (result.get_package_array ());
            if (was_empty && updates_number != 0U) {
                string title = ngettext ("Update Available", "Updates Available", updates_number);
                string body = ngettext ("%u update is available for your system", "%u updates are available for your system", updates_number).printf (updates_number);
                var notification = new Notification (title);
                notification.set_body (body);
                notification.set_icon (new ThemedIcon ("system-software-install"));
                notification.set_default_action ("app.open-application");
                Application.get_default ().send_notification ("updates", notification);
            } else {
                Application.get_default ().withdraw_notification ("updates");
            }

#if HAVE_UNITY
            var launcher_entry = Unity.LauncherEntry.get_for_desktop_file ("org.pantheon.appcenter.desktop");
            launcher_entry.count = updates_number;
            launcher_entry.count_visible = updates_number != 0U;
#endif
        } catch (Error e) {
            critical (e.message);
        }

        updating_cache = false;
        refresh_in_progress = false;
    }

    public uint get_package_count (GLib.GenericArray<weak Pk.Package> package_array) {
        bool os_update_found = false;
        var result_comp = new Gee.TreeSet<AppStream.Component> ();
        package_array.foreach ((pk_package) => {
            var comp = package_list.get (pk_package.get_name ());
            if (comp != null) {
                result_comp.add (comp.component);
            } else {
                os_update_found = true;
            }
        });

        uint size = result_comp.size;
        if (os_update_found) {
            size++;
        }

        return size;
    }

    public void cancel_updates (bool cancel_timeout) {
        interface_cancellable.cancel ();

        if (update_cache_timeout_id > 0 && cancel_timeout) {
            Source.remove (update_cache_timeout_id);
            update_cache_timeout_id = 0;
            last_cache_update = null;
        }

        interface_cancellable = new GLib.Cancellable ();
        refresh_in_progress = false;
    }

    public async void update_cache (bool force = false) {
        debug ("update cache called %s", force.to_string ());
        bool success = false;

        /* Make sure only one update cache can run at a time */
        if (refresh_in_progress) {
            debug ("Update cache already in progress - returning");
            return;
        } else {
            refresh_in_progress = true;
        }


        if (update_cache_timeout_id > 0) {
            if (force) {
                debug ("Forced update_cache called when there is an on-going timeout - cancelling timeout");
                Source.remove (update_cache_timeout_id);
                update_cache_timeout_id = 0;
            } else {
                debug ("Refresh timeout running and not forced - returning");
                refresh_in_progress = false;
                return;
            }
        }

        /* One cache update a day, keeps the doctor away! */
        if (force || last_cache_update == null ||
            (new DateTime.now_local ()).difference (last_cache_update) / GLib.TimeSpan.SECOND >= SECONDS_BETWEEN_REFRESHES) {

            var nm = NetworkMonitor.get_default ();

            if (nm.get_network_available()) {
                debug ("New refresh task");
                var refresh_task = new AppCenter.Task ();


                try {
                    Pk.Results result = yield refresh_task.refresh_cache_async (false, interface_cancellable, (t, p) => { });
                    success = result.get_exit_code () == Pk.Exit.SUCCESS;
                    last_cache_update = new DateTime.now_local ();
                } catch (Error e) {
                    critical ("Update_cache: Refesh cache async failed - %s", e.message);

                }

                if (success) {
                    refresh_updates.begin ();
                }

            } else {
                refresh_in_progress = false; //Stops new timeout while no network.
            }
        } else {
            debug ("Too soon to refresh and not forced");
        }


        if (refresh_in_progress) {
            update_cache_timeout_id = GLib.Timeout.add_seconds (SECONDS_BETWEEN_REFRESHES, () => {
                update_cache_timeout_id = 0;
                update_cache.begin (true);
                return GLib.Source.REMOVE;
            });

            refresh_in_progress = success;
        } // Otherwise updates and timeout were cancelled during refresh, or no network present.
    }

    public async Gee.TreeSet<Pk.Package> get_installed_packages () {
        var packages_task = request_task ();
        var filter = Pk.Bitfield.from_enums (Pk.Filter.INSTALLED, Pk.Filter.NEWEST);
        var installed = new Gee.TreeSet<Pk.Package> ();
        try {
            Pk.Results result = yield packages_task.get_packages_async (filter, null, (prog, type) => {});
            result.get_package_array ().foreach ((pk_package) => {
                installed.add (pk_package);
            });

            release_task (packages_task);
        } catch (Error e) {
            critical (e.message);
            release_task (packages_task);
        }

        return installed;
    }

    private static GLib.Once<Client> instance;
    public static unowned Client get_default () {
        return instance.once (() => { return new Client (); });
    }
}
